# UE5 Fundamentals of Algorithms
### Bachelor of Science in Data Science for Responsible Business

Instructor: [Romain Vuillemot](romain.vuillemot@ec-lyon.fr)

<div style="text-align: center;">
    <img src="figures/logo-ecl.png" style="width:20%; display:inline-block; vertical-align:middle;">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;
    <img src="figures/logo-emlyon.png" style="width:20%; display:inline-block; vertical-align:middle;">
</div>

## Course description

- Basis of algorithms (sorting, search) and data structure (arrays, lists)
- Justification of the choice of data structures
- Calculate the complexity of an algorithm
- Optimize algorithms
- Write advanced programs using efficient algorithms and data structures

## Course material

- [lectures](https://gitlab.ec-lyon.fr/rvuillem/algo-bsc/-/tree/main/lectures): lectures presented in class

- [notebooks](https://gitlab.ec-lyon.fr/rvuillem/algo-bsc/-/tree/main/notebooks): notebook of the lectures with editable code

- [📝 exercises](https://gitlab.ec-lyon.fr/rvuillem/algo-bsc/-/tree/main/exercises): lab exercices

- [solutions](https://gitlab.ec-lyon.fr/rvuillem/algo-bsc/-/tree/main/solutions): solutions of lab exercises

- [📖 books](https://gitlab.ec-lyon.fr/rvuillem/algo-bsc/-/tree/main/books): books and ressources to prepare the lecture & learn more


## Course outline and reading list

### [Lecture 1 - **Data structures and complexity**](lectures/01-data-structures-complexity%20slides.pdf)

- Introduction to data structures
- Asymptotic analysis (Big O notation)
- Data structures
- Trade-offs between time/space complexity and different data structures
- Empirical calculation of complexity

<details>
<summary>📖 </summary>

- [Python for Everybody](books/pythonlearn.pdf) chapter 9 (dictionnaries), chapter 10 (tuples)

</details>

### [Lab 1 - **Data structures and complexity**](exercices/02-recurion-exercises.ipynb)

---

### Lecture 2 - **Recursion and lists**

- Definition and principles of recursion
- Base cases and recursive cases
- Recursion vs. iteration
- Tail recursion
- Lists, Search
- Introduction to lists (arrays vs. linked lists)
- Operations on lists (insertion, deletion, traversal)
- Searching algorithms (linear, binary)

<details>
<summary>📖 </summary>
- [Think Python](books/thinkpython2.pdf) chapter 11 (lists), chapter 12 (tuples)
- [Python for Everybody](books/pythonlearn.pdf) chapter 8 (lists) and Real Python [lists](https://realpython.com/python-list/) 
- [sorts](https://realpython.com/sorting-algorithms-python/).
- [Think Python](books/thinkpython2.pdf) chapter 5 (Conditionals and recursion) and Real Python [recursion](https://realpython.com/python-recursion/).
</details>

### Lab 2 - **Recursion and lists**

---
### Assignment 1
---

### Lecture 3 - **Stacks and queues** 

- Definition and characteristics of stacks and queues
- Stack operations (push, pop, peek)
- Queue operations (enqueue, dequeue, peek)
- Applications

<details>
<summary>📖 </summary>

- 📖 Chapter 1.2.1 (Queue/Stacks), [Open Data Structures](pdf/ods-python.pdf) 
- 📝 Quizz 1

</details>



### Lab 3 - **Stacks and queues** 

---

### Lecture 4 - **Priority queues** 

- Definition and characteristics of priority queues
- Comparison with regular queues (FIFO) and stacks (LIFO)
- Min/Max-priority queues
- Binary heaps

<details>
<summary>📖 </summary>
</details>

### Lab 4 - **Priority queues** 

---

### Lecture 5 - **Programming strategies: divide and conquer, greedy** 

- Recursive, independent problem-solving strategy
- Application to sorting (merge sort, quick sort) and search (binary search)
- Examples: Fibonacci
- Knapsack Problem (greedy)

<details>
<summary>📖 </summary>

- [Divide & conquer](https://en.wikibooks.org/wiki/Algorithms/Divide_and_Conquer)

- [Greedy algorithms](https://en.wikibooks.org/wiki/Algorithms/Greedy_Algorithms)
</details>

### Lab 5 - **Programming strategies: divide and conquer, greedy** 

---

### Lecture 6 - **Programming strategies: dynamic programming**

- Definition and characteristics of dynamic programming
- Memoization data structures
- Examples (using dynamic programming): Fibonacci, Knapsack Problem 


<details>
<summary>📖 </summary>
- [Dynamic programing](https://en.wikibooks.org/wiki/Algorithms/Dynamic_Programming)
</details>


### Lab 6 - **Programming strategies: dynamic programming**

---

### Lecture 7 - **Binary trees**

- Data structures (arrays, adjacency lists)
- Converting between tree representations
- Tree traversal techniques (preorder, inorder, postorder, level-order)

<details>
<summary>📖 </summary>

- Chapter I.3 (Binary Search Trees), [Data Structures and Algorithms](pdf/Dsa.pdf)
- Video on [Binary Tree](https://www.youtube.com/watch?v=pkYVOmU3MgA&t=11510s)

</details>


### Lab 7 - **Binary trees**

---

### Lecture 8 - **Trees**

- Data structures (arrays, adjacency lists)
- Converting between tree representations
- Tree traversal techniques (preorder, inorder, postorder, level-order)

<details>
<summary>📖 </summary>

- Chapter 6 (Trees), [Open Data Structures](pdf/ods-python.pdf)  
- [Problem Solving with Algorithms](pdf/problemsolving.pdf) chapter 6 (trees and tree algorithms)
- [Data Structures and Algorithms](pdf/Dsa.pdf) chapter 13 (binary trees)
 
</details>


### Lab 8 - **Trees**


---

### Lecture 9 - **Graphs**

   - Introduction to graphs (vertices and edges)
   - Types of graphs (directed, undirected, weighted, unweighted)
   - Representing graphs as adjacency matrix, adjacency list
   - Graph properties (connectivity, cycles)

<details>
<summary>📖 </summary>

- [Problem Solving with Algorithms](pdf/problemsolving.pdf) chapter 6.7 (trees traversal)

- [Data Structures and Algorithms](pdf/Dsa.pdf) chapter 12 (introduction to graphs)

- [Data Structures and Algorithms](pdf/Dsa.pdf) chapter 14 (traversing trees)

</details>

### Lab 9

---

### Lecture 10 - **Graphs algorithms (1/2)**

- Minimum spanning tree algorithms (Prim & Kruskal algorithms)
- Topological sorting
- Graph traversal algorithms (depth-first search, breadth-first search)

<details>
<summary>📖 </summary>

- ..

</details>

### Lab 10

---

### Lecture 11 - **Graphs algorithms (2/2)**

- Shortest path algorithms (Bellman-Ford & Dijkstra algorithms)

<details>
<summary>📖 </summary>

- ..

</details>

### Lab 11

---



### Exam

- Final written exam on paper
- 2h
- No document

## Books and ressources

- [Think Python](pdf/thinkpython2.pdf), 2nd edition, by Allen B. Downey
- [Python for Everybody](pdf/pythonlearn.pdf), by Charles Severance
- [Problem Solving with Algorithms and Data Structures using Python](pdf/problemsolving.pdf), by Brad Miller and David Ranum
- [ODS Pyhon](pdf/ods-python.pdf), by Pat Morin ([url](https://opendatastructures.org/ods-python/))

Exercices to practice:
- https://www.w3schools.com/python/exercise.asp
- https://www.practicepython.org/
- https://www.hackerrank.com/domains/python

Other ressources:

- https://en.wikibooks.org/wiki/Algorithms/
- List of [computer science books](https://freecomputerbooks.com/compscAlgorithmBooks.html)